use std::collections::HashMap;

use organic::types::Element;
use organic::types::Object;

type IdCounter = u16;

pub(crate) struct Registry<'intermediate, 'parse> {
    id_counter: IdCounter,
    targets: HashMap<&'parse str, String>,
    footnote_ids: Vec<(
        Option<&'parse str>,
        FootnoteDefinitionContents<'intermediate, 'parse>,
    )>,
}

impl<'intermediate, 'parse> Registry<'intermediate, 'parse> {
    pub(crate) fn new() -> Registry<'intermediate, 'parse> {
        Registry {
            id_counter: 0,
            targets: HashMap::new(),
            footnote_ids: Vec::new(),
        }
    }

    pub(crate) fn get_target<'b>(&'b mut self, body: &'parse str) -> &'b String {
        self.targets.entry(body).or_insert_with(|| {
            self.id_counter += 1;
            format!("target_{}", self.id_counter)
        })
    }

    /// Get a 0-indexed ID for a footnote.
    ///
    /// This needs to be incremented to be 1-indexed for render.
    pub(crate) fn get_footnote_reference_id<'b>(
        &'b mut self,
        label: Option<&'parse str>,
        definition: &'intermediate Vec<Object<'parse>>,
    ) -> usize {
        if let None = label {
            // If it has no label then it must always get a new ID.
            self.footnote_ids
                .push((None, FootnoteDefinitionContents::FromReference(definition)));
            return self.footnote_ids.len() - 1;
        }

        if let Some(existing_id) = self
            .footnote_ids
            .iter()
            .position(|(id, _definition)| *id == label)
        {
            if !definition.is_empty() {
                let entry = self
                    .footnote_ids
                    .get_mut(existing_id)
                    .expect("If-statement proves this to be Some.");
                entry.1 = FootnoteDefinitionContents::FromReference(definition);
            }
            existing_id
        } else {
            self.footnote_ids
                .push((label, FootnoteDefinitionContents::FromReference(definition)));
            self.footnote_ids.len() - 1
        }
    }

    /// Update the definition to a footnote but do not mark it as referenced.
    pub(crate) fn register_footnote_definition<'b>(
        &'b mut self,
        label: &'parse str,
        definition: &'intermediate Vec<Element<'parse>>,
    ) {
        if let Some((existing_id, existing_definition)) = self
            .footnote_ids
            .iter_mut()
            .find(|(id, _definition)| *id == Some(label))
        {
            *existing_definition = FootnoteDefinitionContents::FromDefinition(definition);
        }
    }
}

#[derive(Debug)]
pub(crate) enum FootnoteDefinitionContents<'intermediate, 'parse> {
    FromReference(&'intermediate Vec<Object<'parse>>),
    FromDefinition(&'intermediate Vec<Element<'parse>>),
}
